## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . extension import *
from . vmod import *
from . buildfile import *
from . moduledata import *
## END_IMPORT

class Merger:
    def __init__(self,outname,*inputs):
        self._outName = outname  # Output name 
        self._inputs  = list(inputs)   # File-like objects 
        self._names   = [i.name for i in self._inputs]

        # Pop the first input, and open it as a module, and clone it 
        first     = self._inputs.pop(0)
        firstname = first.name
        first.close()
        with VMod(firstname, 'r') as vfirst:
            self._dest = vfirst.clone(outname,
                                      mode   = 'a',
                                      filter = self.filterCruft)

        self._mergedData = {}
        self._buildFile  = BuildFile(self._dest.getBuildFile())
        self._moduleData = ModuleData(self._dest.getModuleData())
        self._extensions = None
        self._mergedData[firstname] = self._moduleData._root
        self._buildFileDest = (VMod.BUILD_FILE if
                               VMod.BUILD_FILE in self._dest._vmod.namelist()
                               else VMod.BUILD_FILE_SANS)
        
    # ----------------------------------------------------------------
    def filterCruft(self,f):
        '''Filter special.  Return true for files that should be
        filtered out.

        Some cruft left behind by MacOSX is explicitly
        removed, and wrongly embedded module files are also
        filtered out.

        '''                
        if f.startswith('__MACOSX'):
            return True
        if f.endswith('.vmod'):
            return True
        if f.endswith('.DS_Store'):
            return True
        return False
        
    # ----------------------------------------------------------------
    def run(self,overwrite=True,assume_same=False,patch=None):
        with VerboseGuard(f'Merging inputs into {self._outName}') as v:
            try:
                while self._inputs:
                    input             = self._inputs.pop(0)
                    self._currentName = input.name
                    input.close()

                    self.mergeOne(overwrite=True,assume_same=assume_same)
            except:
                raise

            #v(f'{"\n".join(self._dest.getFileNames())}')
            self.documentMerge()
            self.patch(patch)
            
            v(f'Overrding updated files {self._buildFileDest}')
            self._dest.replaceFiles(**{self._buildFileDest:
                                       self._buildFile.encode(),
                                       VMod.MODULE_DATA:
                                       self._moduleData.encode()})

            self.summary()
            
        self._dest._vmod.close()

    # ----------------------------------------------------------------
    def documentMerge(self):
        with VerboseGuard(f'Writing summary of merge') as v:
            doc = self._buildFile.getGame().getDocumentation(single=True)
            if not doc:
                doc = self._buildFile.getGame().addDocumentation()
            else:
                doc = doc[0]
            
            def li(n,d):
                return f'<li><code>{n}</code></li>'

            desc = f'''<html><body>
            <h1>Merged modules and extensions</h1>
            <p>
            This module was created from other modules or extensions by
            the Python script <code>vslmerge.py</code> available from
            </p>
            <pre>
            htps://gitlab.com/wargames_tex/pywargame
            </pre>
            <h1>Merged files</h1>
            <ul>
            {"\n".join([li(n,d) for n,d in self._mergedData.items()])}
            </ul>
            <p>
            See also the XML file <code>merged</code>, in the module
            archive, for more information of the merged modules or
            extensions
            </p>
            </body></html>;'''
            self._dest.addFile('help/merged.html',desc)
            doc.addHelpFile(title='Merged',fileName='help/merged.html')
            v(f'Wrote merge information to help menu')
        
    # ----------------------------------------------------------------
    def summary(self):
        with VerboseGuard(f'Writing summary XML to output') as v:
            from pprint import pprint
            #from xml.dom.minidom import Document
            from pathlib import Path
            
            doc = xmlns.Document()
            lst = doc.createElement('MergeSummary')
            doc.appendChild(lst)
            for n,d in self._mergedData.items():
                p = Path(n).name
                v(f'{p}')
                
                m = doc.createElement('Merged')
                m.setAttribute('filename',p)
                lst.appendChild(m)
                m.appendChild(d.firstChild)

            self._dest.addFile('merged',doc.toprettyxml())


        
            
            
        
    # ----------------------------------------------------------------
    def patch(self,patch):
        if not patch:
            return

        with VerboseGuard(f'Patching merged module w/file {patch.name}') as v:
            from importlib.util import spec_from_file_location, \
                module_from_spec
            from pathlib import Path
            from sys import modules
        
            p = Path(patch.name)
        

            spec   = spec_from_file_location(p.stem, p.absolute())
            module = module_from_spec(spec)
            spec.loader.exec_module(module)
            
            modules[p.stem] = module
            
            module.patch(self._buildFile,
                         self._moduleData,
                         self._dest,
                         Verbose())


    # ----------------------------------------------------------------
    def mergeOne(self,overwrite=True,assume_same=False):
        with VerboseGuard(f'Merging from {self._currentName}') as v:
            with VMod(self._currentName,'r') as self._current:

                if self._current.isExtension():
                    self._mergedData[self._currentName] = \
                        self._current.getExtensionData()
                else:
                    self._mergedData[self._currentName] = \
                        self._current.getModuleData()

                v(f'About to merge files')
                self.mergeFiles(overwrite=overwrite)

                if self._current.isExtension():
                    v(f'About to merge extension data')
                    self.mergeExtension(overwrite=overwrite)
                else:
                    v(f'About to merge module data')
                    self.mergeModuleData(overwrite=overwrite)

                    v(f'About to merge build files')
                    self.mergeBuildFile(overwrite=overwrite,
                                        assume_same=assume_same)
                
    # ----------------------------------------------------------------
    def mergeFiles(self,overwrite=True):
        specials     = [VMod.BUILD_FILE,
                        VMod.BUILD_FILE_SANS,
                        VMod.MODULE_DATA,
                        VMod.EXTENSION_DATA]

        def filter(f):
            return self.filterCruft(f) or f in specials
        
        with VerboseGuard(f'Merging files from {self._currentName}') as v:
            
                
            currentFiles = {f for f in self._current.getFileNames()
                            if not filter(f)}
            destFiles    = {f for f in self._dest.getFileNames()
                            if not filter(f)}
            unique       = currentFiles - destFiles

            v(f'Will add {len(unique)} unique files from '
              f'{self._currentName} to destination')
            uniqueContent = self._current.getFiles(*unique)

            # Do not break line
            v(f'{f"\n{v.i}".join([f"{f}: {len(c)}" for f,c in uniqueContent.items()])}')
            self._dest.addFiles(**uniqueContent)

            if not overwrite:
                return 

            overlap = currentFiles.intersection(destFiles)
            v(f'Will overwrite {len(overlap)} files in destination '
              f'with files from {self._currentName}')

            
            overlapContent = self._current.getFiles(*overlap)
            # Do not break line 
            v(f'{f"\n{v.i}".join([f"{f}: {c[:20]}..." for f,c in overlapContent.items()])}')
            self._dest.replaceFiles(**overlapContent)
            
    # ----------------------------------------------------------------
    def mergeModuleData(self,overwrite=True):
        ''' Merge module data XML''';
        with VerboseGuard(f'Merging module data from '
                          f'{self._currentName}') as v:
            currentModuleDoc  = self._current.getModuleData()
            currentModuleData = ModuleData(currentModuleDoc)

            # v(f'{currentModuleDoc.toprettyxml()}')
            self.mergeElement(self._moduleData,
                              currentModuleData,
                              overwrite=overwrite)

    # ----------------------------------------------------------------
    def mergeBuildFile(self,
                       overwrite   = True,
                       assume_same = False):
        ''' Merge module data XML''';
        with VerboseGuard(f'Merging build file from '
                          f'{self._currentName}') as v:
            currentModuleDoc  = self._current.getBuildFile()
            currentBuildFile  = BuildFile(currentModuleDoc)

            if assume_same:
                with VerboseGuard(f'Assuming same game') as v:
                    # We assume that the modules are modules of the
                    # same game. Thus, we want to change the name of
                    # the top-level element of subsequent modules to
                    # be the same as the destination name.
                    destGame = self._buildFile .getGame()
                    srcGame  = currentBuildFile.getGame()
                    v(f'Destination game is {destGame["name"]}')
                    v(f'Current source game is {srcGame["name"]}')
                    if v: destGame.print()
                    if v: srcGame .print()
                    srcGame['name'] = destGame['name']
                
            # v(f'{currentModuleDoc.toprettyxml()}')
            self.mergeElement(self._buildFile,
                              currentBuildFile,
                              depth       = 0,
                              overwrite   = overwrite)

    # ----------------------------------------------------------------
    def mergeExtension(self,overwrite=True):
        ''' Merge module data XML''';
        with VerboseGuard(f'Merging extension from '
                          f'{self._currentName}') as v:
            currentModuleDoc  = self._current.getBuildFile()
            currentBuildFile  = Extension(parent=currentModuleDoc,
                                          node=currentModuleDoc.firstChild)

            if not self._extensions:
                self._extensions = self._buildFile.getGame().addFolder(
                    name = 'MergedExtensions',
                    description = 'Extensions merged in')

            # Mark extension as loaded
            #
            # Doesn't work because VASSAL.build.module.ModuleExtension
            # does not have a default CTOR - sigh!  Probably done on
            # purpose, bit still annoying.
            #
            # oext = self._extensions.add(Extension)
            # oext.setAttributes(
            #     anyModule         = currentBuildFile['anyModule'],
            #     version           = currentBuildFile['version'],
            #     description       = currentBuildFile['description'],
            #     module            = currentBuildFile['module'],
            #     moduleVersion     = currentBuildFile['moduleVersion'],
            #     vassalVersion     = currentBuildFile['vassalVersion'],
            #     nextPieceSlotId   = currentBuildFile['nextPieceSlotId'],
            #     extensionId       = currentBuildFile['extensionId'])
                

            # Get all the extension elements specified 
            for k,ext in currentBuildFile.getExtensionElements().items():
                v(f'Extension element: {k}/{ext}')
                cur  = self._buildFile
                if ext.target == '':
                    print(f'Warning, no target specified for extension '
                          f'element, will assume top')
                    spec = [[cur._node.firstChild.tagName]]
                else:
                    spec = ext.getSelect()

                # From the unpacked target path, find the target
                # element in destination.
                for tn,*en in spec:
                    v(f' Looking for element w/tag={tn} and attributes={en}')
                    
                    cs = Element.getTagClass(tn)
                    if not cs:
                        raise RuntimeError(f'Got no class for tag={tn}')

                    es = cur.getAllElements(cs,single=len(en)<1)
                    if not es:
                        raise RuntimeError(f'Got no elements w/tag={tn}')

                    if en:
                        tgt = None
                        unt = Element._make_unique(tn,*en)
                        for e in es:
                            #v(f'  candidate: {e._unique()} ?= {unt}')
                            if e._unique() == unt:
                                tgt = e
                                break
                    else:
                        tgt = es[0]
                            

                    if not tgt:
                        raise RuntimeError(f'Failed to find element w/tag='
                                           f'{tn} and "name"={en}')
                        
                    cur = tgt

                if not cur:
                    raise RuntimeError(f'Failed to find element w/tag='
                                       f'{tn} and "name"={en}')

                # We have our target element.  As we cannot specify
                # changed attributes in extension elements, all we
                # need to do is to merge in the child elements 
                v(f'Target element is {cur}')

                self.mergeElement(cur,
                                  ext,
                                  depth          = 0,
                                  overwrite      = overwrite,
                                  skipAttributes = True)

                
            
            
    # ----------------------------------------------------------------
    def mergeElement(self,
                     dest,
                     src,
                     depth          = 0,
                     overwrite      = True,
                     skipAttributes = False):
        '''Merge element src into dest under the policy 'overwrite'

        This is a multi-stage process.

        First, we check if there are attributes to merge, and then do that.
        Second, we check if there are child nodes to merge, and then do that. 
        '''
        with VerboseGuard(f'Merging element '
                          f'{src._unique()}') as v:
            if not skipAttributes:
                self.mergeAttributes(dest,src,overwrite)
            self.mergeChildren (dest,src,
                                depth       = depth,
                                overwrite   = overwrite)

            if overwrite and dest.hasText() and src.hasText():
                # Perhaps append?
                dest.setText(src.getText())
                #dest.setText(dest.getText()+' '+src.getText())


    # ----------------------------------------------------------------
    def mergeAttributes(self,dest,src,overwrite=True):
        with VerboseGuard(f'Merging attributes of elements '
                          f'{src._unique()}') as v:
            srcAttributes  = src.getAttributes()
            destAttributes = dest.getAttributes()
            srcNames       = set(srcAttributes .keys()
                                 if srcAttributes  else [])
            
            if srcAttributes is None:
                return
            
            destNames      = set(destAttributes .keys()
                                 if destAttributes  else [])
            unique         = srcNames - destNames

            v(f'Adding attributes {unique}')
            dest.setAttributes(**{k:v for k,v in srcAttributes.items()
                                  if k in unique})

            if not overwrite:
                return

            overlap        = srcNames.intersection(destNames)
            v(f'Overwriting attributes {overlap}')
            dest.setAttributes(**{k:v for k,v in srcAttributes.items()
                                  if k in overlap})
        
    # ----------------------------------------------------------------
    def mergeChildren(self,dest,src,
                      overwrite   = True,
                      depth       = 0):
        if depth > 20:
            print(f'Maximum depth {depth} reached')
            return
        
        with VerboseGuard(f'Merging children of element '
                          f'{src._unique()}') as v:
            srcChildren  = set(src. getAllElements(cls=None))
            destChildren = set(dest.getAllElements(cls=None))
            unique       = srcChildren - destChildren
            overlap      = destChildren.intersection(srcChildren)

            with VerboseGuard(f'Adding {len(unique)} unique elements from '
                              f'{self._currentName} to destination {dest}') as vv:
                for e in unique:
                    vv(f'Adding unique element {e._unique()}')
                    dest.append(e)
                    # print(dest._node.toprettyxml())

            with VerboseGuard(f'Merging {len(overlap)} children from '
                              f'{self._currentName} into destination') as vv:
                for e in overlap:
                    ds = [ee for ee in destChildren if ee == e]
                    if len(ds) != 1:
                        raise RuntimeError('Should not happen')
                    d = ds[0]
                    # Only apply `assume_same` on top-level
                    self.mergeElement(d,e,
                                      depth       = depth+1,
                                      overwrite   = overwrite)

#
# EOF
#



#!/usr/bin/env python
## BEGIN_IMPORTS
from exporter    import GBXExporter
from common      import Verbose
from gamebox     import GameBox
from extractor   import GBXExtractor
## END_IMPORTS

# ====================================================================
if __name__ == '__main__':
    from argparse import ArgumentParser, FileType

    ap = ArgumentParser(description='Create draft VASSAL module')
    ap.add_argument('gbxfile',
                    help='The GBX file to data from',
                    type=FileType('r'))
    ap.add_argument('-p','--patch',
                    help='A python script to patch generated module',
                    type=FileType('r'))
    ap.add_argument('-o','--output',
                    help='Output file to write module to',
                    type=FileType('w'),
                    default='Draft.vmod')
    ap.add_argument('-r','--rules',
                    help='Rules PDF file',
                    type=FileType('r'))
    ap.add_argument('-T','--tutorial',
                    help='Tutorial (v)log file',
                    type=FileType('r'))
    ap.add_argument('-W','--vassal-version',
                    help='Vassal version number',
                    type=str,
                    default='3.6.7')
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')
    ap.add_argument('-G','--visible-grids',
                    action='store_true',
                    help='Make grids visible in the module')


    args = ap.parse_args()

    gbxname   = args.gbxfile.name
    vmodname  = args.output.name
    rulesname = args.rules.name    if args.rules    is not None else None
    tutname   = args.tutorial.name if args.tutorial is not None else None
    args.output.close()
    args.gbxfile.close()

    patchname = args.patch.name if args.patch is not None else None
    if args.patch is not None:
        args.patch.close()

    Verbose().setVerbose(args.verbose)
        
    try:
        gamebox   = GameBox.fromFile(gbxname)
        extractor = GBXExtractor(gamebox)
        exporter  = GBXExporter(extractor._d,
                                rules         = rulesname,
                                tutorial      = tutname,
                                visible       = args.visible_grids,
                                vassalVersion = args.vassal_version)
        exporter.run(vmodname,patchname)
    except Exception as e:
        from sys import stderr 
        print(f'Failed to build {vmodname}: {e}',file=stderr)
        from os import unlink
        try:
            unlink(vmodname)
        except:
            pass

        raise e




## BEGIN_IMPORT
from .. common import VerboseGuard
from . image import GBXImage
from . base import CbManager
from . features import Features
## END_IMPORT

# ====================================================================
class GBXTileLocation:
    def __init__(self,ar):
        '''Where a tile can be found'''
        with VerboseGuard('Reading tile location') as g:
            self._sheet  = (ar.word() if not Features().size_size == 8 else
                            ar.size())
            self._offset = ar.word() 
            g(f'Tile location at sheet={self._sheet} offset={self._offset}')
            if self._sheet == 65535: self._sheet = -1

    def __str__(self):
        return f'{self._sheet:3d} @ {self._offset:6d}'

        
# --------------------------------------------------------------------
class GBXTileDefinition:
    def __init__(self,ar):
        '''The definition of a tile'''
        with VerboseGuard('Reading tile definition'):
            self._full = GBXTileLocation(ar)
            self._half = GBXTileLocation(ar)
            self._fill = ar.dword()

    def __str__(self):
        return f'Full: {self._full}, Half: {self._half}, Fill: {self._fill:08x}'
    
        
# --------------------------------------------------------------------
class GBXTileSet:
    def __init__(self,ar):
        '''A set of tiles'''
        with VerboseGuard('Reading tile set'):
            self._name = ar.str()
            n = ar.word() if Features().size_size != 8 else ar.size()
            self._ids  = [ar.iden()
                          for _ in range(n)]

    def __str__(self):
        return (self._name + ':' + ','.join([str(i) for i in self._ids]))

    
# --------------------------------------------------------------------
class GBXTileSheet:
    def __init__(self,ar,transparent):
        '''A single image that has multiple tile images in it
        
        X,Y are the tile sizes
        '''
        with VerboseGuard('Reading tile sheet'):
            self._x      = ar.word()
            self._y      = ar.word()
            hasBM        = ar.word()
            self._img    = GBXImage(ar,transparent) if hasBM else None

    def sub(self,off):
        if self._img is None:
            return None

        return self._img._img.crop((0,off,self._x,off+self._y))
    
    def __str__(self):
        bm = str(self._img) if self._img is not None else 'None'
        return (f'c=({self._x:4d},{self._y:4d}) bitmap={bm}')
        

# --------------------------------------------------------------------
class GBXTileManager(CbManager):
    def __init__(self,ar):
        '''Tile manager (including tiles)'''
        with VerboseGuard('Reading tile manager'):
            self._transparent = ar.dword()
            super(GBXTileManager,self).__init__(ar)
            
            ts = lambda ar : GBXTileSheet(ar, self._transparent)
            self._tiledefs    = self._readSub(ar,GBXTileDefinition,
                                              Features().id_size)
            self._tilesets    = self._readSub(ar,GBXTileSet)
            self._tilesheets  = self._readSub(ar,ts)
            self._toStore     = {} # Used in boards, not elsewhere

    def image(self,tileID):
        if tileID is None:
            return None 
        if tileID == 0xFFFF:
            return None

        tileDef = self._tiledefs[tileID]
        tileSht = self._tilesheets[tileDef._full._sheet]
        img     = tileSht.sub(tileDef._full._offset)
        return img

    def store(self,tileID):
        filename = self._toStore.get(tileID,{}).get('filename',None)
        if filename is None:
            filename              = f'tile_{tileID:04d}.png'
            self._toStore[tileID] = {
                'filename': filename,
                'image'   : self.image(tileID)
            }

        return filename
        
        
    def __str__(self):
        return ('Tile manager:\n'
                 + f'  Transparent:  {self._transparent:08x}\n'
                 + super(GBXTileManager,self).__str__()
                 + self._strSub('tiles',self._tiledefs) + '\n'
                 + self._strSub('tile sets',self._tilesets) + '\n'
                 + self._strSub('tile sheets',self._tilesheets))

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . head import num_version, readVector
from . features import Features
## END_IMPORT
    
class GSNTraySet:
    def __init__(self,ar,vers):
        with VerboseGuard(f'Reading tray set') as g:
            self._name          = ar.str()
            g(f'Tray set: {self._name}')
            self._random        = (ar.word() if Features().piece_100 else 0)
            self._visibleFlags  = ar.dword()
            self._ownerMask     = (ar.word() if vers < num_version(3,10) else
                                   ar.dword())
            self._restrict      = ar.word()
            self._pieces        = readVector(ar,lambda ar: ar.iden())

    def __len__(self):
        return len(self._pieces)

    def toDict(self):
        viz = self._visibleFlags & ~0xFFFF8000
        # print(f'{self._visibleFlags:08x} -> {viz}')
        vizStr = {0: 'all',
                  1: 'owner',
                  2: 'generic',
                  3: 'none'}.get(viz,'')
        return {'name':     self._name,
                'visible':  vizStr,
                'owner':    self._ownerMask,
                'restrict': self._restrict,
                'pieces':   self._pieces }
    
    def __str__(self):
        return (f'Tray set: {self._name} '
                f'[visible={self._visibleFlags},'
                f'ownerMask={self._ownerMask},'
                f'resrict={self._restrict}] '
                f'({len(self)}): {self._pieces}')


class GSNTrayManager:
    def __init__(self,ar,vers,iden=''):
        with VerboseGuard(f'Reading tray {iden} manager @ {ar.tell()}') as g:
            self._iden     = iden
            self._reserved = [ar.word() for _ in range(4)]
            g(f'{self._reserved}')
            self._dummy    = (ar.byte() if vers >= num_version(4,0) else 0)
            self._sets     = readVector(ar, lambda ar : GSNTraySet(ar,vers))

    def __len__(self):
        return len(self._sets)

    def toDict(self):
        return [s.toDict() for s in self._sets]
    
    def __str__(self):
        pl = '\n    '.join([str(s) for s in self._sets])
        return f'TrayManager: {self._iden} ({len(self)})\n    {pl}\n'


#
# EOF
#

        
            
            



#!/usr/bin/env python
#
# Patch a module
## BEGIN_IMPORTS
from pywargame.vassal import VMod
from pywargame.common import Verbose
## END_IMPORT

# ====================================================================
def patchIt(args):
    Verbose().setVerbose(args.verbose)
    
    vmodname  = args.output.name
    patchname = args.patch.name
    args.output.close()
    args.patch .close()
    
    VMod.patch(vmodname, patchname, args.verbose)

# ====================================================================
def patchMain():
    from argparse import ArgumentParser, FileType

    ap = ArgumentParser(description='Patch a module with a Python script')
    ap.add_argument('output',
                    help='Module to patch, will be overwritten',
                    type=FileType('r'))
    ap.add_argument('patch',
                    help='Python script to patch module',
                    type=FileType('r'),
                    nargs='?',
                    default='patch.py')
    ap.add_argument('-V','--verbose',
                    help='Be verbose',
                    action='store_true')

    args = ap.parse_args()
    
    patchIt(args)
    
# ====================================================================
if __name__ == '__main__':
    patchMain()
    
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . withtraits import *
from . globalkey import *
## END_IMPORT

# --------------------------------------------------------------------
class MapElementService:
    def getMap(self):
        '''Get map - either a Map or WidgetMap'''
        ## BEGIN_IMPORT
        from . map import WidgetMap, Map
        ## END_IMPORT
        return self.getParentOfClass([WidgetMap,Map])
        # if self._parent is None:
        #     return None
        # 
        # if 'WidgetMap' in self._parent.tagName:
        #     return self.getParent(WidgetMap)
        #     
        # return self.getParent(Map)
    def getGame(self):
        m = self.getMap()
        if m is not None: return m.getGame()
        return None

# --------------------------------------------------------------------
class MapElement(Element,MapElementService):
    def __init__(self,map,tag,node=None,**kwargs):
        super(MapElement,self).__init__(map,tag,node=node,**kwargs)


# --------------------------------------------------------------------
class PieceLayers(MapElement):
    TAG=Element.MAP+'LayeredPieceCollection'
    UNIQUE = ['property']
    def __init__(self,map,node=None,
                 property = 'PieceLayer',
                 description = '',
                 layerOrder = []):
        super(PieceLayers,self).__init__(map,self.TAG,node=node,
                                         property    = property,
                                         description = description,
                                         layerOrder  = ','.join(layerOrder))

    def addControl(self,**kwargs):
        '''Add `LayerControl` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : LayerControl
            The added element
        '''
        return self.add(LayerControl,**kwargs)
    def getControls(self,asdict=True):
        '''Get all `LayerControl` element(s) from this

        Parameters
        ----------
        asdict : bool        
            If `True`, return a dictonary that maps name to
            `LayerControl` elements.  If `False`, return a list of all
            `LayerControl` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `LayerControl` children

        '''
        return self.getElementsByKey(LayerControl,'name',asdict)
                 
registerElement(PieceLayers)
    
# --------------------------------------------------------------------
class LayerControl(MapElement):
    TAG=Element.MAP+'LayerControl'
    CYCLE_UP='Rotate Layer Order Up'
    CYCLE_DOWN='Rotate Layer Order Down'
    ENABLE='Make Layer Active'
    DISABLE='Make Layer Inactive'
    TOGGLE='Switch Layer between Active and Inactive'
    RESET='Reset All Layers'
    UNIQUE = ['name']
    def __init__(self,col,node=None,
                 name         = '',
                 tooltip      = '',
                 text         = '',
                 hotkey       = '',
                 icon         = '',
                 canDisable   = False,
                 propertyGate = '', #Property name, disable when property false
                 disabledIcon = '',
                 command      = TOGGLE,
                 skip         = False,
                 layers       = [],
                 description = ''):
        super(LayerControl,self).__init__(col,self.TAG,node=node,
                                          name         = name,
                                          tooltip      = tooltip,
                                          text         = text,
                                          buttonText   = text,
                                          hotkey       = hotkey,
                                          icon         = icon,
                                          canDisable   = canDisable,
                                          propertyGate = propertyGate,
                                          disabledIcon = disabledIcon,
                                          command      = command,
                                          skip         = skip,
                                          layers       = ','.join(layers),
                                          description  = description)

    def getLayers(self):
        '''Get map - either a Map or WidgetMap'''
        return self.getParentOfClass([PieceLayers])
        
registerElement(LayerControl)
        

# --------------------------------------------------------------------
class LineOfSight(MapElement):
    TAG=Element.MAP+'LOS_Thread'
    ROUND_UP        = 'Up'
    ROUND_DOWN      = 'Down'
    ROUND_NEAREST   = 'Nearest whole number'
    FROM_LOCATION   = 'FromLocation'
    TO_LOCATION     = 'ToLocation'
    CHECK_COUNT     = 'NumberOfLocationsChecked'
    CHECK_LIST      = 'AllLocationsChecked'
    RANGE           = 'Range'
    NEVER           = 'Never'
    ALWAYS          = 'Always'
    WHEN_PERSISTENT = 'When persistent'
    CTRL_CLICK      = 'Cltr-Click & Drag'    
    UNIQUE          = ['threadName']
    
    def __init__(self,map,
                 node=None,
                 threadName         = 'LOS',
                 hotkey             = key('L',ALT),
                 tooltip            = 'Trace line of sight',
                 iconName           = '/images/thread.gif', #'los-icon.png',
                 label              = '',
                 snapLOS            = False,
                 snapStart          = True,
                 snapEnd            = True,
                 report             = (f'{{"Range from "+{FROM_LOCATION}'
                                       f'+" to "+{TO_LOCATION}+" is "'
                                       f'+{RANGE}+" (via "+{CHECK_LIST}+")"}}'),
                 persistent         = CTRL_CLICK,
                 persistentIconName = '/images/thread.gif',
                 globl              = ALWAYS,
                 losThickness       = 3,
                 threadColor        = rgb(255,0,0),
                 drawRange          = True,
                 # rangeBg            = rgb(255,255,255),
                 # rangeFg            = rgb(0,0,0),
                 rangeScale         = 0,
                 hideCounters       = True,
                 hideOpacity        = 50,
                 round              = ROUND_UP,
                 canDisable         = False,
                 propertyGate       = '',
                 disabledIcon       = ''):
        '''Make Line of Sight interface
        
        Parameters
        ----------
        threadName : str
            Name of interface
        hotkey : str
            Start LOS key
        tooltip : str
            Tool tip text
        iconName : str
            Path to button icon
        label : str
            Button text 
        snapLOS : bool
            Wether to snap both ends
        snapStart : bool
            Snap to start
        snapEnd: bool
            Snap to end
        report : str
            Report format
        persistent : str
            When persistent
        persistentIconName : str
            Icon when persistent(?)
        globl : str
            Visisble to opponents
        losThickness : int
            Thickness in pixels
        losColor : str
            Colour of line
        drawRange : bool
            Draw the range next to LOST thread
        rangeBg : str
            Range backgroung colour
        rangeFg : str
            Range foregrond colour
        rangeScale : int
            Scale of range - pixels per unit
        round : str
            How to round range
        hideCounters :bool
            If true, hide counters while making thread
        hideOpacity : int
            Opacity of hidden counters (percent)
        canDisable : bool
            IF true, then can be hidden
        propertyGate : str
            Name of property.  When that property is TRUE, then the
            interface is disabled.  Must be a property name, not an expression.
        disabledIcon : str
            Icon to use when disabled
        '''
        super(LineOfSight,self).__init__(map,self.TAG,
                                         node = node,
                                         threadName         = threadName,
                                         hotkey             = hotkey,
                                         tooltip            = tooltip,
                                         iconName           = iconName,
                                         label              = label,
                                         snapLOS            = snapLOS,
                                         snapStart          = snapStart,
                                         snapEnd            = snapEnd,
                                         report             = report,
                                         persistent         = persistent,
                                         persistentIconName = persistentIconName,
                                         losThickness       = losThickness,
                                         threadColor        = threadColor,
                                         drawRange          = drawRange,
                                         #rangeBg            = rangeBg,
                                         #rangeFg            = rangeFg,
                                         rangeScale         = rangeScale,
                                         hideCounters       = hideCounters,
                                         hideOpacity        = hideOpacity,
                                         round              = round,
                                         canDisable         = canDisable,
                                         propertyGate       = propertyGate,
                                         disabledIcon       = disabledIcon)
        self.setAttribute('global',globl)
                                     
    
registerElement(LineOfSight)
    
# --------------------------------------------------------------------
class StackMetrics(MapElement):
    TAG=Element.MAP+'StackMetrics'
    def __init__(self,map,node=None,
                 bottom               = key('(',0),
                 down                 = key('%',0),
                 top                  = key('&',0),
                 up                   = key("'",0),
                 disabled             = False,
                 exSepX               = 16,  # Expanded (after double click)
                 exSepY               = 24,  # Expanded (after double click)
                 unexSepX             = 8,   # Compact
                 unexSepY             = 16): # Compact
        super(StackMetrics,self).__init__(map,self.TAG,node=node,
                                          bottom               = bottom,
                                          disabled             = disabled,
                                          down                 = down,
                                          exSepX               = exSepX,
                                          exSepY               = exSepY,
                                          top                  = top,
                                          unexSepX             = unexSepX,
                                          unexSepY             = unexSepY,
                                          up                   = up)

registerElement(StackMetrics)

# --------------------------------------------------------------------
class MoveCamera(MapElement):
    TAG=Element.MAP+'MoveCameraButton'
    LOCATION = 'L'  # Pixel location
    ZONE     = 'Z'  # Zone on board
    REGION   = 'R'  # Region on board
    GRID     = 'G'  # Grid location 
    COUNTER  = 'A'  # Counter by property
    CYCLE    = 'C'  # Cycle through counters by property
    NEAREST  = 'N'  # Nearest counter by property
    def __init__(self,map,node=None,
                 name             = '',            # Comment
                 tooltip          = 'Move camera to location',
                 text             = 'Move camera', # Button text
                 hotkey           = key('C',ALT),  # Hot key
                 icon             = 'recenter.gif',# Button icon
                 canDisable       = False,         # Can be disabled
                 propertyGate     = '',            # Disable when property true
                 disabledIcon     = '',            # Hidden button icon
                 hideWhenDisabled = False,         # Do not show when hidden
                 zoom             = 0,             # Adjust zoom - 0: no chang 
                 moveCameraMode   = GRID,          # Where to move camera
                 boardName        = '',            # Board to move to
                 xPos             = 0,             # Pixel X
                 yPos             = 0,             # Pixel Y
                 zoneName         = '',            # Zone name
                 gridLocation     = '',            # Grid Location
                 regionName       = '',            # Region name
                 propertyFilter   = '',            # Piece 
                 xOffset          = 0,             # Additional offset
                 yOffset          = 0):            # Additional offset
        super(MoveCamera,self).__init__(map,self.TAG,node=node,
                                        name             = name,
                                        tooltip          = tooltip,
                                        text             = text,
                                        hotkey           = hotkey,
                                        icon             = icon,
                                        canDisable       = canDisable,
                                        propertyGate     = propertyGate,
                                        disabledIcon     = disabledIcon,
                                        hideWhenDisabled = hideWhenDisabled,
                                        zoom             = zoom,
                                        moveCameraMode   = moveCameraMode,
                                        boardName        = boardName,
                                        xPos             = xPos,
                                        yPos             = yPos,
                                        zoneName         = zoneName,
                                        gridLocation     = gridLocation,
                                        regionName       = regionName,
                                        propertyFilter   = propertyFilter,
                                        xOffset          = xOffset,
                                        yOffset          = yOffset)

registerElement(MoveCamera)

# --------------------------------------------------------------------
class ImageSaver(MapElement):
    TAG=Element.MAP+'ImageSaver'
    def __init__(self,map,node=None,
                 buttonText           = '',
                 canDisable           = False,
                 hotkey               = '',
                 icon                 = '/images/camera.gif',
                 propertyGate         = '',
                 tooltip              = 'Save map as PNG image'):
        super(ImageSaver,self).__init__(map,self.TAG,node=node,
                                        buttonText           = buttonText,
                                        canDisable           = canDisable,
                                        hotkey               = hotkey,
                                        icon                 = icon,
                                        propertyGate         = propertyGate,
                                        tooltip              = tooltip)

registerElement(ImageSaver)

# --------------------------------------------------------------------
class TextSaver(MapElement):
    TAG=Element.MAP+'TextSaver'
    def __init__(self,map,node=None,
                 buttonText           = '',
                 canDisable           = False,
                 hotkey               = '',
                 icon                 = '/images/camera.gif',
                 propertyGate         = '',
                 tooltip              = 'Save map as text'):
        super(TextSaver,self).__init__(map,self.TAG,node=node,
                                        buttonText           = buttonText,
                                        canDisable           = canDisable,
                                        hotkey               = hotkey,
                                        icon                 = icon,
                                        propertyGate         = propertyGate,
                                        tooltip              = tooltip)

registerElement(TextSaver)

# --------------------------------------------------------------------
class ForwardToChatter(MapElement):
    TAG=Element.MAP+'ForwardToChatter'
    def __init__(self,map,node=None,**kwargs):
        super(ForwardToChatter,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(ForwardToChatter)

# --------------------------------------------------------------------
class MenuDisplayer(MapElement):
    TAG=Element.MAP+'MenuDisplayer'
    def __init__(self,map,node=None,**kwargs):
        super(MenuDisplayer,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(MenuDisplayer)

# --------------------------------------------------------------------
class MapCenterer(MapElement):
    TAG=Element.MAP+'MapCenterer'
    def __init__(self,map,node=None,**kwargs):
        super(MapCenterer,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(MapCenterer)

# --------------------------------------------------------------------
class StackExpander(MapElement):
    TAG=Element.MAP+'StackExpander'
    def __init__(self,map,node=None,**kwargs):
        super(StackExpander,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(StackExpander)

# --------------------------------------------------------------------
class PieceMover(MapElement):
    TAG=Element.MAP+'PieceMover'
    def __init__(self,map,node=None,**kwargs):
        super(PieceMover,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(PieceMover)

# --------------------------------------------------------------------
class SelectionHighlighters(MapElement):
    TAG=Element.MAP+'SelectionHighlighters'
    def __init__(self,map,node=None,**kwargs):
        super(SelectionHighlighters,self).\
            __init__(map,self.TAG,node=node,**kwargs)

registerElement(SelectionHighlighters)

# --------------------------------------------------------------------
class KeyBufferer(MapElement):
    TAG=Element.MAP+'KeyBufferer'
    def __init__(self,map,node=None,**kwargs):
        super(KeyBufferer,self).__init__(map,self.TAG,node=node,**kwargs)

registerElement(KeyBufferer)

# --------------------------------------------------------------------
class HighlightLastMoved(MapElement):
    TAG=Element.MAP+'HighlightLastMoved'
    def __init__(self,map,node=None,
                 color     = rgb(255,0,0),
                 enabled   = True,
                 thickness = 2):
        super(HighlightLastMoved,self).__init__(map,self.TAG,node=node,
                                                color     = color,
                                                enabled   = enabled,
                                                thickness = thickness)

registerElement(HighlightLastMoved)

# --------------------------------------------------------------------
class CounterDetailViewer(MapElement):
    TAG=Element.MAP+'CounterDetailViewer'
    TOP_LAYER  = 'from top-most layer only'
    ALL_LAYERS = 'from all layers'
    INC_LAYERS = 'from listed layers only'
    EXC_LAYERS = 'from layers other than those listed'
    FILTER     = 'by using a property filter'
    ALWAYS     = 'always'
    NEVER      = 'never'
    IF_ONE     = 'ifOne'
    UNIQUE     = ['description']
    def __init__(self,map,node=None,
                 borderWidth            = 0, # Horizontal padding between pieces
                 borderThickness        = 2, # Outer border thickness
                 borderInnerThickness   = 2, # Inner borders thickness
                 borderColor            = None,
                 centerAll              = False,
                 centerText             = False,
                 centerPiecesVertically = True,
                 combineCounterSummary  = False,
                 counterReportFormat    = '',
                 delay                  = 700,
                 description            = '',
                 display                = TOP_LAYER,
                 emptyHexReportForma    = '$LocationName$',
                 enableHTML             = True,
                 extraTextPadding       = 0,
                 bgColor                = None,
                 fgColor                = rgb(0,0,0),
                 fontSize               = 11,
                 graphicsZoom           = 1.0,# Zoom on counters
                 hotkey                 = key('\n'),
                 layerList              = [],
                 minDisplayPieces       = 2,
                 propertyFilter         = '',
                 showDeck               = False,
                 showDeckDepth          = 1,
                 showDeckMasked         = False,
                 showMoveSelected       = False,
                 showNoStack            = False,
                 showNonMovable         = False,
                 showOverlap            = False,
                 showgraph              = True,
                 showgraphsingle        = False,
                 showtext               = True,
                 showtextsingle         = False,
                 stretchWidthSummary    = False,
                 summaryReportFormat    = '$LocationName$',
                 unrotatePieces         = False,
                 version                = 4,
                 verticalOffset         = 2,
                 verticalTopText        = 0,
                 zoomlevel              = 1.0,
                 stopAfterShowing       = False,
                 showTerrainBeneath     = NEVER,
                 showTerrainSnappy      = True,
                 showTerrainWidth       = 120,
                 showTerrainHeight      = 120,
                 showTerrainZoom        = None,
                 showTerrainText        = ''
                 ): # showTerrain attributes

        bg = '' if bgColor is None else bgColor
        fg = '' if fgColor is None else fgColor
        bc = '' if borderColor is None else borderColor
        ll = ','.join(layerList)
        showTerrainZoom = zoomlevel if showTerrainZoom == None else showTerrainZoom
        super(CounterDetailViewer,self)\
            .__init__(map,self.TAG,node=node,
                      borderWidth            = borderWidth,
                      borderThickness        = borderThickness,
                      borderInnerThickness   = borderInnerThickness,
                      borderColor            = bc,
                      centerAll              = centerAll,
                      centerText             = centerText,
                      centerPiecesVertically = centerPiecesVertically,
                      combineCounterSummary = combineCounterSummary,
                      counterReportFormat   = counterReportFormat,
                      delay                 = delay,
                      description           = description,
                      display               = display, # How to show from layers
                      emptyHexReportForma   = emptyHexReportForma,
                      enableHTML            = enableHTML,
                      extraTextPadding      = extraTextPadding,
                      bgColor               = bg,
                      fgColor               = fg,
                      fontSize              = fontSize,
                      graphicsZoom          = graphicsZoom, # pieces at zoom
                      hotkey                = hotkey,
                      layerList             = ll,
                      minDisplayPieces      = minDisplayPieces,
                      propertyFilter        = propertyFilter,
                      showDeck              = showDeck,
                      showDeckDepth         = showDeckDepth,
                      showDeckMasked        = showDeckMasked,
                      showMoveSelectde      = showMoveSelected,
                      showNoStack           = showNoStack,
                      showNonMovable        = showNonMovable,
                      showOverlap           = showOverlap,
                      showgraph             = showgraph,
                      showgraphsingle       = showgraphsingle,
                      showtext              = showtext,
                      showtextsingle        = showtextsingle,
                      stretchWidthSummary   = stretchWidthSummary,
                      summaryReportFormat   = summaryReportFormat,
                      unrotatePieces        = unrotatePieces,
                      version               = version,
                      verticalOffset        = verticalOffset,
                      verticalTopText       = verticalTopText,
                      zoomlevel             = zoomlevel,
                      stopAfterShowing      = stopAfterShowing,
                      showTerrainBeneath    = showTerrainBeneath,
                      showTerrainSnappy     = showTerrainSnappy,
                      showTerrainWidth      = showTerrainWidth,
                      showTerrainHeight     = showTerrainHeight,
                      showTerrainZoom       = showTerrainZoom,
                      showTerrainText       = showTerrainText)

registerElement(CounterDetailViewer)

# --------------------------------------------------------------------
class GlobalMap(MapElement):
    TAG=Element.MAP+'GlobalMap'
    def __init__(self,map,node=None,
                 buttonText           = '',
                 color                = rgb(255,0,0),
                 hotkey               = key('O',CTRL_SHIFT),
                 icon                 = '/images/overview.gif',
                 scale                = 0.2,
                 tooltip              = 'Show/Hide overview window'):
        super(GlobalMap,self).\
            __init__(map,self.TAG,node=node,
                     buttonText           = buttonText,
                     color                = color,
                     hotkey               = hotkey,
                     icon                 = icon,
                     scale                = scale,
                     tooltip              = 'Show/Hide overview window')

registerElement(GlobalMap)

# --------------------------------------------------------------------
class Zoomer(MapElement):
    TAG = Element.MAP+'Zoomer'
    def __init__(self,map,node=None,
                 inButtonText         = '',
                 inIconName           = '/images/zoomIn.gif',
                 inTooltip            = 'Zoom in',
                 outButtonText        = '',
                 outIconName          = '/images/zoomOut.gif',
                 outTooltip           = 'Zoom out',
                 pickButtonText       = '',
                 pickIconName         = '/images/zoom.png',
                 pickTooltip          = 'Select Zoom',
                 zoomInKey            = key('=',CTRL_SHIFT),
                 zoomLevels           = [0.2,0.25,0.333,0.4,0.5,
                                         0.555,0.625,0.75,1.0,1.25,1.6],
                 zoomOutKey           = key('-'),
                 zoomPickKey          = key('='),
                 zoomStart            = 3):

        '''Zoom start is counting from the back (with default zoom levels,
        and zoom start, the default zoom is 1'''
        lvls = ','.join([str(z) for z in zoomLevels])
        super(Zoomer,self).\
            __init__(map,self.TAG,node=node,
                     inButtonText         = inButtonText,
                     inIconName           = inIconName,
                     inTooltip            = inTooltip,
                     outButtonText        = outButtonText,
                     outIconName          = outIconName,
                     outTooltip           = outTooltip,
                     pickButtonText       = pickButtonText,
                     pickIconName         = pickIconName,
                     pickTooltip          = pickTooltip,
                     zoomInKey            = zoomInKey,
                     zoomLevels           = lvls,
                     zoomOutKey           = zoomOutKey,
                     zoomPickKey          = zoomPickKey,
                     zoomStart            = zoomStart)

registerElement(Zoomer)

# --------------------------------------------------------------------
class HidePiecesButton(MapElement):
    TAG=Element.MAP+'HidePiecesButton'
    def __init__(self,map,node=None,
                 buttonText           = '',
                 hiddenIcon           = '/images/globe_selected.gif',
                 hotkey               = key('O'),
                 showingIcon          = '/images/globe_unselected.gif',
                 tooltip              = 'Hide all pieces on this map'):
        super(HidePiecesButton,self).\
            __init__(map,self.TAG,node=node,
                     buttonText           = buttonText,
                     hiddenIcon           = hiddenIcon,
                     hotkey               = hotkey,
                     showingIcon          = showingIcon,
                     tooltip              = tooltip)
        
registerElement(HidePiecesButton)

# --------------------------------------------------------------------
class MassKey(GlobalKey,MapElementService):
    TAG = Element.MAP+'MassKeyCommand'
    UNIQUE     = ['name']
    def __init__(self,map,node=None,
                 name                 = '',                
                 buttonHotkey         = '',
                 hotkey               = '',
                 buttonText           = '',
                 canDisable           = False,
                 deckCount            = '-1',
                 filter               = '',
                 propertyGate         = '',
                 reportFormat         = '',
                 reportSingle         = False,
                 singleMap            = True,
                 target               = GlobalKey.SELECTED,
                 tooltip              = '',
                 icon                 = ''):
        '''Default targets are selected units'''
        super(MassKey,self).\
            __init__(map,self.TAG,node=node,
                     name                 = name,                
                     buttonHotkey         = buttonHotkey, # This hot key
                     hotkey               = hotkey,       # Target hot key
                     buttonText           = buttonText,
                     canDisable           = canDisable,
                     deckCount            = deckCount,
                     filter               = filter,
                     propertyGate         = propertyGate,
                     reportFormat         = reportFormat,
                     reportSingle         = reportSingle,
                     singleMap            = singleMap,
                     target               = target,
                     tooltip              = tooltip,
                     icon                 = icon)

registerElement(MassKey)

# --------------------------------------------------------------------
class Flare(MapElement):
    TAG=Element.MAP+'Flare'
    def __init__(self,map,node=None,
                 circleColor          = rgb(255,0,0),
                 circleScale          = True,
                 circleSize           = 100,
                 flareKey             = 'keyAlt',
                 flareName            = 'Map Flare',
                 flarePulses          = 6,
                 flarePulsesPerSec    = 3,
                 reportFormat         = ''):
        super(Flare,self).__init__(map,self.TAG,node=node,
                                   circleColor          = circleColor,
                                   circleScale          = circleScale,
                                   circleSize           = circleSize,
                                   flareKey             = flareKey,
                                   flareName            = flareName,
                                   flarePulses          = flarePulses,
                                   flarePulsesPerSec    = flarePulsesPerSec,
                                   reportFormat         = '')

registerElement(Flare)

# --------------------------------------------------------------------
class Deck(MapElement):
    TAG = Element.MODULE+'map.DrawPile'
    ALWAYS     = 'Always'
    NEVER      = 'Never',
    VIA_MOUSE2 = 'Via right-click Menu'
    UNIQUE     = ['name','owningBoard']
    def __init__(self,map,
                 node                  = None,
                 name                  = 'deckName',
                 owningBoard           = '',
                 x                     = 0,   # int
                 y                     = 0,   # int
                 width                 = 200, # int
                 height                = 200, # int
                 #
                 allowMultiple         = False,
                 drawMultipleMessage   = 'Draw multiple cards',
                 #
                 allowSelect           = False,
                 drawSpecificMessage   = 'Draw specific cards',
                 selectDisplayProperty = '$BasicName$',
                 selectSortProperty    = 'BasicName',
                 #
                 faceDown              = ALWAYS,#ALWAYS,VIA_MOUSE2
                 faceFlipHotkey        = key('F'),
                 faceDownFormat        = '',
                 faceDownHotkey        = '',
                 faceDownMessage       = 'Face down',
                 faceUpHotkey          = '',
                 faceUpMessage         = 'Face up',
                 faceUpReportFormat    = '',
                 drawFaceUp            = False,
                 #
                 shuffle               = VIA_MOUSE2,#ALWAYS,NEVER
                 shuffleCommand        = 'Shuffle',
                 shuffleFormat         = '$playerSide$ shuffles $deckName$',
                 shuffleHotkey         = key('S',ALT),
                 #
                 reversible            = False,
                 reverseCommand        = 'Reverse',
                 reverseFormat         = '',
                 reverseHotkey         = '',
                 #
                 draw                  = True,
                 color                 = rgb(255,51,51),
                 hotkeyOnEmpty         = False,
                 emptyHotkey           = key(NONE,0)+',DeckEmpty',
                 #
                 reshufflable          = False,
                 reshuffleCommand      = '',
                 reshuffleHotkey       = '',
                 reshuffleMessage      = '',
                 reshuffleTarget       = '',
                 #
                 canSave               = False,
                 saveHotkey            = '',
                 saveMessage           = 'Save Deck',
                 saveReportFormat      = 'Deck Saved',
                 loadHotkey            = '',
                 loadMessage           = 'Load Deck',
                 loadReportFormat      = 'Deck Loaded',
                 #
                 maxStack              = 15,
                 #
                 expressionCounting    = False,
                 countExpressions      = '',
                 #
                 restrictExpression    = '',
                 restrictOption        = False,
                 #
                 deckOwners            = '',
                 deckRestrictAccess    = False
                 ): # int
        pass
        super(Deck,self).\
            __init__(map,self.TAG,node=node,
                     name                  = name,
                     owningBoard           = owningBoard,
                     x                     = int(x),      # int
                     y                     = int(y),      # int
                     width                 = int(width),  # int
                     height                = int(height), # int
                     #
                     allowMultiple         = allowMultiple,
                     drawMultipleMessage   = drawMultipleMessage,
                     #
                     allowSelect           = allowSelect,
                     drawSpecificMessage   = drawSpecificMessage,
                     selectDisplayProperty = selectDisplayProperty,
                     selectSortProperty    = selectSortProperty,
                     #
                     faceDown              = faceDown,
                     faceFlipHotkey        = faceFlipHotkey,
                     faceDownFormat        = faceDownFormat,
                     faceDownHotkey        = faceDownHotkey,
                     faceDownMessage       = faceDownMessage,
                     faceUpHotkey          = faceUpHotkey,
                     faceUpMessage         = faceUpMessage,
                     faceUpReportFormat    = faceUpReportFormat,
                     drawFaceUp            = drawFaceUp,
                     #
                     shuffle               = shuffle,
                     shuffleCommand        = shuffleCommand,
                     shuffleFormat         = shuffleFormat,
                     shuffleHotkey         = shuffleHotkey,
                     #
                     reversible            = reversible,
                     reverseCommand        = reverseCommand,
                     reverseFormat         = reverseFormat,
                     reverseHotkey         = reverseHotkey,
                     #
                     draw                  = draw,
                     color                 = color,
                     hotkeyOnEmpty         = hotkeyOnEmpty,
                     emptyHotkey           = emptyHotkey,
                     #
                     reshufflable          = reshufflable,
                     reshuffleCommand      = reshuffleCommand,
                     reshuffleHotkey       = reshuffleHotkey,
                     reshuffleMessage      = reshuffleMessage,
                     reshuffleTarget       = reshuffleTarget,
                     #
                     canSave               = canSave,
                     saveHotkey            = saveHotkey,
                     saveMessage           = saveMessage,
                     saveReportFormat      = saveReportFormat,
                     loadHotkey            = loadHotkey,
                     loadMessage           = loadMessage,
                     loadReportFormat      = loadReportFormat,
                     #
                     maxStack              = maxStack,
                     #
                     expressionCounting    = expressionCounting,
                     countExpressions      = countExpressions,
                     #
                     restrictExpression    = restrictExpression,
                     restrictOption        = restrictOption,
                     #
                     deckOwners            = deckOwners,
                     deckRestrictAccess    = deckRestrictAccess
                     )

    def addCard(self,**kwargs):
        '''Add a `Card` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Card
            The added element
        '''
        if not isinstance(card,CardSlot):
            print(f'Trying to add {type(card)} to Deck')
            return None
            
        p = card.clone(self)
        # self._node.appendChild(p._node)
        return p
    def addFolder(self,**kwargs):
        '''Add a `ModuleFolder` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : ModuleFolder
            The added element
        '''
        return self.add(DeckFolder,**kwargs)
        
    def getCards(self,asdict=True):
        '''Get all Card element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Card`
            elements.  If `False`, return a list of all Card`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `Card` children

        '''
        return self.getElementsByKey(CardSlot,'entryName',asdict)
    def getFolders(self,asdict=True):
        '''Get all Menu element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Folder`
            elements.  If `False`, return a list of all `Folder`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `Folder` children

        '''
        return self.getElementsByKey(DeckFolder,'name',asdict)

    
registerElement(Deck)
        
                 
# --------------------------------------------------------------------
class AtStart(MapElement):
    TAG = Element.MODULE+'map.SetupStack'
    UNIQUE     = ['name','location','owningBoard']
    def __init__(self,map,
                 node            = None,
                 name            = '',
                 location        = '',
                 useGridLocation = True,
                 owningBoard     = '',
                 x               = 0,
                 y               = 0):
        '''Pieces are existing PieceSlot elements


        Parameters
        ----------
        node : xml.minidom.Node
            Existing node or None
        name : str
            Name of node
        location : str
            Where the at-start element is put if `useGridLocation`
        useGridLocation : bool
            If true, use maps grid
        owningBoard : str
            Board that owns the at-start (can be empty)
        x : float
            Coordinate (ignored if `useGridLocation`)
        y : float
            Coordinate (ignored if `useGridLocation`)
        '''
        super(AtStart,self).\
            __init__(map,self.TAG,node=node,
                     name            = name,
                     location        = location,
                     owningBoard     = owningBoard,
                     useGridLocation = useGridLocation,
                     x               = x,
                     y               = y)

    def addPieces(self,*pieces):
        '''Add a `Pieces` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Pieces
            The added element
        '''
        # copy pieces here
        copies = []
        for p in pieces:
            c = self.addPiece(p)
            if c is not None:
                copies.append(c)
        return copies
        
    def addPiece(self,piece):
        '''Add a `Piece` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Piece
            The added element
        '''
        if not isinstance(piece,WithTraitsSlot):
            # Next is a bit of a hack - not nice
            if piece.__class__.__name__ not in ['PieceSlot','CardSlot']:
                print(f'Trying to add {type(piece)} to AtStart, '
                      f'not a {isinstance(piece,WithTraitsSlot)}')
                return None
            
        p = piece.clone(self)
        # self._node.appendChild(p._node)
        return p
    
    def getPieces(self,asdict=True):
        '''Get all Piece element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Piece`
            elements.  If `False`, return a list of all Piece`
            children.

        Returns
        -------
        children : dict or list
            Dictionary or list of `Piece` children

        '''
        return self.getElementsByKey(PieceSlot,'entryName',asdict)

registerElement(AtStart)

# --------------------------------------------------------------------
class ForwardKeys(MapElement):
    TAG = Element.MODULE+'map.ForwardToKeyBuffer'
    def __init__(self,map,
                 node            = None):
        '''Forward keys to key buffer from where it is distributed to
        selected pieces.  Don't know how I missed this!

        ''' 
        
        super(ForwardKeys,self).\
            __init__(map,self.TAG,node=node)

registerElement(ForwardKeys)

# --------------------------------------------------------------------
class Scroller(MapElement):
    TAG = Element.MODULE+'map.Scroller'
    ALWAYS = 'always'
    NEVER  = 'never'
    PROMPT = 'prompt'
    def __init__(self,map,
                 node           = None,
                 useArrows      = PROMPT):
        '''This component listens to key events on a Map window and
        scrolls the map.  Depending on the useArrows attribute, will
        use number keypad or arrow keys, or will offer a preferences
        setting for the user to choose
        ''' 
        
        super(Scroller,self).\
            __init__(map,self.TAG,node=node,
                     useArrows = useArrows)

registerElement(Scroller)

#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class StackTrait(Trait):
    ID = 'stack'
    def __init__(self,
                 board     = '',
                 x         = '',  
                 y         = '',  
                 pieceIds  = [],
                 layer     = -1): 
        '''Create a stack trait in a save file'''
        self.setType()       # NAME
        # print('Piece IDs:',pieceIds)
        self.setState(board      = board,
                      x          = x,
                      y          = y,
                      pieceIds   = ';'.join([str(p) for p in pieceIds]),
                      layer      = f'@@{layer}')
        
Trait.known_traits.append(StackTrait)

#
# EOF
#



# ====================================================================
class Singleton(type):
    '''Meta base class for singletons'''
    _instances = {}
    def __call__(cls, *args, **kwargs):
        '''Create the singleton object or returned existing

        Parameters
        ----------
        args : tuple
            Arguments
        kwargs : dict
            Keyword arguments
        '''
        if cls not in cls._instances:
            cls._instances[cls] = \
                super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

#
# EOF
#



## BEGIN_IMPORTS
from pywargame.common import VerboseGuard, Verbose
from pywargame.vassal.buildfile import BuildFile
from pywargame.vassal.documentation import Documentation
from pywargame.vassal.traits import *
from pywargame.vassal.base import *
from pywargame.vassal.moduledata import ModuleData
from pywargame.vassal.exporter import Exporter
from pywargame.vassal.mapelements import LayerControl
from pywargame.vassal.vsav import VSav
from . base import ZTImage
## END_IMPORTS


class ZTExporter(Exporter):
    def __init__(self,
                 gamebox,
                 version       = None,
                 vassalVersion = '3.6',
                 allmaps       = False,
                 mainMap       = None,
                 nrotations    = 6,
                 symbolic_dice = True):
        self._gamebox       = gamebox
        self._version       = '0.0.1' if version is None else version
        self._vassalVersion = vassalVersion
        self._allmaps       = allmaps
        self._main_map      = mainMap
        self._n_rotations   = nrotations
        self._symbolic_dice = symbolic_dice

    # ----------------------------------------------------------------
    def setup(self):
        with VerboseGuard('Saving images'):
            from pathlib import Path


    # ----------------------------------------------------------------
    def createModuleData(self):
        with VerboseGuard('Creating module data'):
            self._moduleData = ModuleData()
            data             = self._moduleData.addData()
            data.addVersion      (version='0.0.1')
            data.addVASSALVersion(version=self._vassalVersion)
            data.addName         (name   =self._gamebox._name)
            data.addDescription  (description=self._gamebox._desc)
            data.addDateSaved    ()

    # ----------------------------------------------------------------
    def createBuildFile(self,
                        ignores = ['all',
                                   'common',
                                   'marker',
                                   'markers',
                                   ' ']):
        with VerboseGuard(f'Creating build file') as v:
            self._build = BuildFile() # 'buildFile.xml')
            self._game  = self._build.addGame(
                name        = self._gamebox._name,
                version     = self._version,
                description = self._gamebox._desc)
            
            v(f'Adding description')
            self.addDocumentation()
            
            v(f'Adding command encoder')
            self._game.addBasicCommandEncoder()

            v(f'Adding Global options')
            go = self._game.addGlobalOptions(
                autoReport         = 'Use Preferences Setting',
                centerOnMove       = 'Use Preferences Setting',
                nonOwnerUnmaskable = 'Use Preferences Setting',
                playerIdFormat     = '$playerName$')
            go.addOption(name='undoHotKey',value=key('Z'))

            v(f'Adding dice hands')
            self.addDiceHands()
            
            v(f'Adding prototypes')
            self.addPrototypes()

            v(f'Adding pieces')
            self.addPieces()

            v(f'Fixing up map locations')
            self.fixMaps()
            
            v(f'Adding maps')
            self.addMaps()

            v(f'Adding counter sheets')
            self.addCounterSheets()

            v(f'Adding scenarios')
            self.addScenarios()

    def fixMaps(self):
        # Scale all scenarios to target resolution before anything else
        for sc in self._gamebox._scenarios:
            sc.scale(ZTImage.target_dpi / 600)

        for m in self._gamebox._maps:
            self.fixMap(m)

    def fixMap(self,map):
        from wand.image import Image, Color
        from wand.drawing import Drawing
        
        with VerboseGuard(f'Fixing up map {map._name}') as v:
            tgt = [s for s in self._gamebox._scenarios if s.on_map(map._name)]
            if not tgt:
                v(f'No target scenario found for map')
                return

            # if len(tgt) > 1:
            #     v(f'More than 1 target scenario found for map')
            #     return
            dxs = []
            dys = []
            ws  = []
            hs  = []
            for sc in tgt:
                v(f'Target scenario for map "{map._name}" is "{sc._name}"')
                v(f'Map "{map._name}" image is '
                  f'{map._image._image.width}x{map._image._image.height}')
            
                ly = sc.map_layout(map._name)[0]
                v(f'Layout at {ly.left},{ly.top} {ly.lrx},{ly.lry}')

                bb  = sc.calc_bounding_box(map._name if self._allmaps
                                           else None)
                v(f'Bounding box for map: {bb}')
                
                ulx = min([s._x - self._gamebox._piece_map[p]._front.width/2
                           for s in ly._stacks for p in s._ids]+[0])
                lrx = max([s._x + self._gamebox._piece_map[p]._front.width/2
                           for s in ly._stacks for p in s._ids]+
                          [map._image._image.width])
                uly = min([s._y - self._gamebox._piece_map[p]._front.height/2
                           for s in ly._stacks for p in s._ids]+[0])
                lry = max([s._y + self._gamebox._piece_map[p]._front.height/2
                           for s in ly._stacks for p in s._ids]+
                          [map._image._image.height])
                v(f'Bounding box for map: {bb} {ulx},{uly},{lrx},{lry}')

                # Translate the layout
                mar = 10
                sdx = -int(ulx-.5)+mar
                sdy = -int(uly-.5)+mar
                sw  = int(lrx+sdx+.5)+2*mar
                sh  = int(lry+sdy+.5)+2*mar
                dxs.append(sdx)
                dys.append(sdy)
                ws .append(sw)
                hs .append(sh)
                v(f'{sc._name}: dx={sdx} dy={sdy} w={sw} h={sh}')
                
            # Find the smallest translation, and largest size
            dx = max(dxs)
            dy = max(dys)
            w  = max(ws)
            h  = max(hs)
            v(f'All scenarios translated by {dx},{dy}')
            for sc in tgt:
                sc.translate(dx,dy)
                # v(f'{sc}')
            
            v(f'Background image {w}x{h}')
            bg = Image(width=w,height=h,
                       colorspace='truecolor',
                       background=Color('#404040'))
            bg.format = 'png'
            #map._image._image.save(filename=map.filename)
            
            v(f'Compose over at {dx},{dy}')
            bg.composite(map._image._image,left=dx,top=dy)

            # Replace with new image
            #bg.save(filename='bg_'+map.filename)
            map._image._image = bg #.save(filename=map.filename)
            
    # ----------------------------------------------------------------
    def saveImages(self,vmod):
        pass

    # ----------------------------------------------------------------
    def savePNG(self,vmod,filename,img):
        #with VerboseGuard(f'Saving PNG {filename}'):
        from io import BytesIO

        stream = BytesIO()
        img.save(stream)
        vmod.addFile('images/'+filename,stream.getvalue())

    # ----------------------------------------------------------------
    def addDocumentation(self):
        '''Add documentation to the module.  This includes rules,
        key-bindings, and about elements.
        '''
        with VerboseGuard('Adding documentation') as v:
            doc = self._game.addDocumentation()

            desc = f'''<html><body>
            <h1>Module</h1>
            <p>
            This module was created from a ZunTzu gamebox
            {self._gamebox._name} by the Python script
            <code>ztexport.py</code> available from
            </p>
            <p>
            Ctrl-click to select terrain pieces.
            </p>
            <pre>
            https://gitlab.com/wargames_tex/pywargames
            </pre>
            <h1>Game</h1>
            <p>{self._gamebox._desc}</p>
            <h1>Scenarios</h1>
            <ul>
            {'\n'.join(['<li>'+s._name+': '+s._desc+'</li>'
                        for s in self._gamebox._scenarios])}            
            </ul>
            <h1>Copyright</h1>
            <p>&copy; {self._gamebox._copy}</p>
            </body></html>;'''

            self._vmod.addFile('help/description.html',desc)
            doc.addHelpFile(title='Description',
                            fileName='help/description.html')

            if self._gamebox._splash:
                v(f'Adding splash screen') 
                fname = 'splash.'+self._gamebox._splash.format.lower()
                self.savePNG(self._vmod,fname,self._gamebox._splash)
                self._gamebox._splash = None
                doc.addAboutScreen(title=self._gamebox._name,
                                   fileName=fname)
                
                
            
    # ----------------------------------------------------------------
    def addDiceHands(self):
        with VerboseGuard(f'Adding dice hands: '
                          f'{self._gamebox._dice_hands}') as v:
            seen = set()
            for no,hand in enumerate(self._gamebox._dice_hands):
                name = f'{len(hand._dice)}d{hand._type}'
                if name in seen:
                    name += f'_{no}'
                    
                seen.add(name)
                # v(f'Prepare to add dice {name}')
                self.addDiceHand(name,hand)

    # ----------------------------------------------------------------
    def addDiceHand(self,name,hand):
        with VerboseGuard(f'Adding dice hand {name} '
                          f'{len(hand._dice)}d{hand._type}') as v:

            cnt  = len(hand._dice)
            sid  = hand._type

            rep = '$name$ = '
            if cnt > 1:
                rep += ' + '.join([f'$result{i+1}$' for i in range(cnt)])+' = '
            rep += '$result$'
            ky  =  {4:  '4',
                    6:  '6',
                    8:  '8',
                    10: '0',
                    12: '2',
                    20: '3'}[hand._type]
            #v(f'Dice defined: {hand._dice}')
            if not self._symbolic_dice:
                self._game.addDiceButton(name         = name,
                                         hotkey       = key(ky,ALT),
                                         tooltip      = f'Roll {cnt}d{sid}',
                                         text         = name,
                                         nDice        = cnt,
                                         nSides       = sid,
                                         reportFormat = rep,
                                         reportTotal  = True)
                return


            # Could be used instead if we knew we had images
            #
            v(f'Adding symbolic dice')

            # Make icon.  A nicer way would be to combine the images
            # created below and then scale to the appropriate height.
            #
            # For another time.
            # drawer = DiceDrawer(hand._type,20,20,
            #                     fg='white',
            #                     bg='black')
            # image = drawer.draw(hand._type//2)
            # image.format = 'png'
            # fn    = f'{name}-logo.png'
            # self.savePNG(self._vmod,fn,image)

            rep = '{name+": "+'
            def img(i,d):
                return (f'"<img src=\'{name}{i}_"+result{i}+".png\' '
                        f'width=\'24\' height=\'24\'>"')
            def txt(i,d):
                bg = d['pips']
                fg = d['color']
                bg = 0xFFFFFF
                if fg == 0xFFFFFF:
                    fg = 0

                return (f'"<span style=\'color:#{fg:06x}; '
                        f'background-color:#{bg:06x}; '
                        f'font-weight:bold; '
                        f'padding: 5px;\'>"+'
                        f'result{i}+"</span>"')
                
            rep += '+'.join([img(i+1,d)
                                 for i,d in enumerate(hand._dice)])+\
                                         '+" = "+'
            rep += '+" + "+'.join([txt(i+1,d)
                                   for i,d in enumerate(hand._dice)])
            if cnt > 1:
                rep += '+" = <b>"+numericalTotal+"</b>"}'
            else:
                rep += '}'
                

            diceW  = 80
            diceH  = 80
            iconfn = f'{name}-icon.png'
            but    = self._game.addSymbolicDice(
                name         = name,
                icon         = iconfn,
                hotkey       = key(str(hand._type),ALT),
                tooltip      = f'Roll {cnt}d{sid}',
                text         = name,
                doReport     = True,
                resultWindow = True,
                format       = rep,
                windowX      = (diceW+10)*len(hand._dice),
                windowY      = diceH+10)
            v(f'Report format: {rep}')

            from wand.image import Image as WImage
            from wand.color import Color
            from random import randint
            
            icon = WImage(format='png',width=diceW,height=diceH)
            icon.alpha_channel = True
            icon.background_color = Color('transparent')
            
            for i, dice in enumerate(hand._dice):
                # v(f'- {dice}')
                dien = f'{name}{i+1}'
                die  = but.addDie(name=dien)
                drawer = DiceDrawer(hand._type,80,80,
                                    fg=dice['pips'],
                                    bg=dice['color'])

                imgs = []
                for num in range(1,hand._type+1):
                    if hand._type == 10 and num == 10:
                        num = 0
                    image = drawer.draw(num)
                    image.format = 'png'
                    imgs.append(image)
                    fn    = f'{dien}_{num}.png'
                    self.savePNG(self._vmod,fn,image)
                    
                    die.addFace(icon = fn,
                                text = f'{num}',
                                value = num)

                icon.sequence.append(imgs[randint(0,hand._type-1)])

            icon.smush(stacked=False,offset=0)
            sc = 20 / diceH
            icon.resize(int(sc * icon.width),int(sc * icon.height))
            icon.format = 'png'
            self.savePNG(self._vmod,iconfn,icon)
            
            
            
    # ----------------------------------------------------------------
    def addPrototypes(self):
        with VerboseGuard(f'Creating prototypes'):
            protos = self._game.addPrototypes()

            traits = [TrailTrait(),
                      RotateTrait(nangles = self._n_rotations),
                      DeleteTrait(),
                      BasicTrait()]
            protos.addPrototype(name        = 'Basic prototype',
                                description = 'Prototype for most',
                                traits      = traits)


            traits = [PrototypeTrait(name      = 'Basic prototype'),
                      NoStackTrait(
                          select      = NoStackTrait.CTRL_SELECT,
                          bandSelect  = NoStackTrait.NEVER_BAND_SELECT,
                          move        = NoStackTrait.SELECT_MOVE,
                          canStack    = False,
                          description = 'Terrain does not stack'),
                      MarkTrait(name  = 'PieceLayer',
                                value = 'Terrains'),
                      BasicTrait()]
            protos.addPrototype(name        = 'Terrain prototype',
                                description = 'Prototype for terrain',
                                traits      = traits)

            traits = [PrototypeTrait(name = 'Basic prototype'),
                      MarkTrait(name  = 'PieceLayer',
                                value = 'Pieces'),
                      BasicTrait()]

            protos.addPrototype(name        = 'Piece prototype',
                                description = 'Prototype for pieces',
                                traits      = traits)

            traits = [PrototypeTrait(name = 'Basic prototype'),
                      MarkTrait(name  = 'PieceLayer',
                                value = 'Cards'),
                      BasicTrait()]

            protos.addPrototype(name        = 'Card prototype',
                                description = 'Prototype for cards',
                                traits      = traits)
            
    # ----------------------------------------------------------------
    def addPieces(self):
        with VerboseGuard(f'Adding pieces') as v:
            window = self._game.addPieceWindow(name = 'Pieces',
                                               hotkey = key('C',ALT))

            combo  = window.addCombo(entryName = 'Pieces')
            
            self._piece_slots = {}

            for sheet in self._gamebox._counter_sheets:
                v(sheet._name)
                if len(sheet._piece) <= 0 and len(sheet._card) <= 0:
                    continue
                plist = self.addPieceContainer(combo, sheet._name)
                self.addSheetPieces(plist,sheet._piece,sheet._card)
            
    # ----------------------------------------------------------------
    def addSheetPieces(self,container,pieces,cards):
        with VerboseGuard(f'Adding pieces to {container["entryName"]} '+
                          f'{len(pieces)} pieces, {len(cards)} cards'):
            for piece in pieces+cards:
                boardName = container['entryName']
                board     = self._game.getBoards(asdict=True).get(boardName,None)
                mapName   = board.getMap()['mapName'] if board else ''
                no     = self._gamebox._piece_id[piece]
                traits = [
                    PrototypeTrait(
                        name = (('Terrain' if piece._terrain else
                                 ('Card' if piece._card else 'Piece'))+
                                ' prototype')),
                    SendtoTrait(
                        mapName     = boardName,
                        boardName   = mapName,
                        name        = f'Return to {boardName}',
                        key         = key('R',CTRL_SHIFT),
                        restoreName = '',
                        restoreKey  = '',
                        x           = piece._x,
                        y           = piece._y,
                        description = 'Return to where it came from',
                        destination = SendtoTrait.LOCATION)
                ]
                name       = f'{no:06d}'
                front_name = f'{name}_front.{piece._front.format.lower()}'
                back_name  = (f'{name}_back.{piece._back.format.lower()}'
                              if piece.two_sides else None)
                if back_name is not None:
                    traits.extend([
                        LayerTrait(
                            images       = [front_name,back_name],
                            newNames     = ['','+ reversed'],
                            activateName = '',
                            decreaseName = '',
                            increaseName = 'Flip',
                            increaseKey  = key('F'),
                            decreaseKey  = '',
                            name         = 'Step'),
                        ReportTrait(key('F'))])
                gpid = self._game.nextPieceSlotId()
                size = piece.size
                traits.append(
                    BasicTrait(name     = name,
                               filename = front_name,
                               gpid     = gpid))
                
                self._piece_slots[piece] = container.addPieceSlot(
                    entryName = name,
                    gpid      = gpid,
                    traits    = traits,
                    width     = piece.width,
                    height    = piece.height)

                for img,fname in zip([piece._front,piece._back],
                                     [front_name,back_name]):
                    if img is None:
                        continue
                    fnameext = fname
                    self.savePNG(self._vmod,fnameext,img)
                #piece._front = None
                #piece._back  = None
                    
            # self.saveImages(self,vmod)
    # ----------------------------------------------------------------
    def addPieceContainer(self,container,name):
        #panel = container.addPanel(entryName = name,
        #                           fixed = False,
        #                           vert  = True)
        #return panel
        panel = container.addPanel(entryName = name, fixed = False)
        plist = panel.addList(entryName = f'{name}',
                              width     = 300,
                              height    = 300,
                              divider   = 150)
        return plist
    
    # ----------------------------------------------------------------
    def addMapLayers(self,map):
        layerDesc = {'Terrains': {'t': 'Terrain', 'i': ''},
                     'Cards':    {'t': 'Cards',   'i': ''},
                     'Pieces':   {'t': 'Pieces',  'i': ''}}
        layers = map.addLayers(property = 'PieceLayer',
                               layerOrder = layerDesc)
        for layerTitle, layerData in layerDesc.items():
            layers.addControl(name    = layerTitle,
                              tooltip = f'Toggle display of {layerTitle}',
                              text    = f'Toggle {layerData["t"]}',
                              icon    = layerData['i'],
                              layers  = [layerTitle])
        layers.addControl(name = 'Show all',
                          tooltip = 'Show all',
                          text    = 'Show all',
                          command = LayerControl.ENABLE,
                          layers  = list(layerDesc.keys()))
        map.addMenu(description = 'Toggle layers',
                    text        = '',
                    tooltip     = 'Toggle display of layers',
                    icon        = '/images/inventory.gif',
                    menuItems   = ([f'Toggle {ln["t"]}'
                                        for ln in layerDesc.values()]
                                       +['Show all']))         

    # ----------------------------------------------------------------
    def addMapDefaults(self,map):
        s      = ZTImage.target_dpi / 150
        z      = [0.05,0.1,0.2,0.25,0.333,0.4,0.5,
                  0.555,0.625,0.75,1.0,1.25,1.6,1.8,2]        
        if s >  4: z = z[:-3]
        if s <  2: z = z[2:]
        z0           = 6
        map.addGlobalMap()
        # Basics
        map.addStackMetrics(
            exSepX    = int(s * 12 + .5),
            exSepY    = int(s * 16 + .5),
            unexSepX  = int(s *  8 + .5),
            unexSepY  = int(s * 10 + .5)
        )
        map.addImageSaver()
        map.addTextSaver()
        map.addForwardToChatter()     
        map.addMenuDisplayer()        
        map.addMapCenterer()          
        map.addStackExpander()        
        map.addPieceMover()           
        map.addKeyBufferer()
        map.addForwardKeys()
        map.addSelectionHighlighters()
        map.addHighlightLastMoved()
        map.addScroller()
        map.addZoomer(zoomLevels = z, zoomStart = z0)
        
    # ----------------------------------------------------------------
    def addMaps(self):
        with VerboseGuard(f'Adding maps') as v:
            if not self._allmaps:
                self.addMap(*self._gamebox._maps)
                return

            mapNames = [self.addMap(m) for m in self._gamebox._maps]

            filtered = [n for n in mapNames if n != '' and n != self._main_map]
            v(f'Filtered set of maps: {filtered}')
            if len(filtered) > 0:
                self._game.addMenu(description = 'Maps',
                                   text        = 'Maps',
                                   tooltip     = 'Show or hide maps',
                                   icon        = '/images/map.gif',
                                   menuItems   = filtered)
                
                
    # ----------------------------------------------------------------
    def addMap(self,*maps):
        with VerboseGuard(f'Adding maps') as v:
            isMain = False
            mapName     = maps[0]._name
            if not self._main_map or self._main_map == mapName:
                self._main_map = mapName
                isMain = True
                
            map = self._game.addMap(mapName    = mapName,
                                    buttonName = '' if isMain else mapName,
                                    hotkey     = '' if isMain else key('M',ALT),
                                    launch     = not isMain
                                    )
            self.addMapDefaults(map)
            map.addLOS()
            map.addCounterDetailViewer()
            self.addMapLayers(map)
            
            picker  = map.addBoardPicker()
            self._picker = picker
            for m in maps:
                imgname = m.filename
                size    = m.size
                brd     = picker.addBoard(name   = m._name,
                                          image  = imgname,
                                          width  = size[0],
                                          height = size[1])
                zoned  = brd.addZonedGrid()
                zoned.addHighlighter()
                zone   = zoned.addZone(name = 'Full',
                                       useParentGrid = False,
                                       path=(f'{0},{0};' +
                                             f'{size[0]},{0};' +
                                             f'{size[0]},{size[1]};' +
                                             f'{0},{size[1]}'))

                self.savePNG(self._vmod,imgname,m._image._image)
                m._image._image = None

            return mapName
    
    # ----------------------------------------------------------------
    def addCounterSheets(self):
        with VerboseGuard(f'Adding Charts') as v:
            window = \
                self._game.addChartWindow(name='Charts',
                                          hotkey = key('A',ALT),
                                          description = 'Charts',
                                          text        = '',
                                          icon       = '/images/chart.gif',
                                          tooltip     = 'Show/hide Charts')
            tabs = window.addTabs(entryName='Charts')

            for sheet in self._gamebox._counter_sheets:
                widget = tabs.addMapWidget(entryName=sheet._name)
                self.addCounterSheet(widget,sheet)

    # ----------------------------------------------------------------
    def addCounterSheet(self,widget,sheet):
        with VerboseGuard(f'Adding counter sheet {sheet._name}') as v:
            map = widget.addWidgetMap(mapName   = sheet._name,
                                      markMoved = 'Never',
                                      hotkey    = '')
            self.addMapDefaults(map)
            self.addMapLayers(map)
            
            size   = sheet.size
            picker = map.addPicker()
            brd    = picker.addBoard(name  = sheet._name,
                                     image = sheet.filename)
            zoned  = brd.addZonedGrid()
            zoned.addHighlighter()
            zone   = zoned.addZone(name = 'Full',
                                   useParentGrid = False,
                                   path=(f'{0},{0};' +
                                         f'{size[0]},{0};' +
                                         f'{size[0]},{size[1]};' +
                                         f'{0},{size[1]}'))

            self.savePNG(self._vmod,sheet.filename,sheet._front._image)
            self.addAtStart(map,sheet._piece+sheet._card)
            sheet._front._image = None
            
    # ----------------------------------------------------------------
    def addAtStart(self,map,pieces):
        with VerboseGuard(f'Adding at-start to {map["mapName"]} '+
                          f'{len(pieces)} pieces and cards') as g:
            toAdd = {}
            
            for piece in pieces:
                x    = piece._x
                y    = piece._y
                grid = (x,y)
                slot = self._piece_slots[piece]
                #g(f'Adding piece at {x},{y}')
                if grid not in toAdd:
                    toAdd[grid] = {'center': (x,y),
                                   'pieces': [] }
                toAdd[grid]['pieces'].append(slot)

            for grid, dpieces in toAdd.items():
                center  = dpieces['center']
                name    = f'{grid[0]:06d}{grid[1]:06d}'
                atstart = map.addAtStart(name            = name,
                                         useGridLocation = False,
                                         owningBoard     = map["mapName"],
                                         x               = center[0],
                                         y               = center[1])
                atstart.addPieces(*dpieces['pieces'])

    # ----------------------------------------------------------------
    def addDeadMap(self,sides):
        pass

    # ----------------------------------------------------------------
    def addScenarios(self):
        menu = self._game.addPredefinedSetup(name = 'Setups',
                                             isMenu = True)

        menu.addPredefinedSetup(name = 'No setup',
                                file = '',
                                useFile = False,
                                description = 'No initial setup')
        
        for sc in self._gamebox._scenarios:
            self.addScenario(sc,menu)

    # ----------------------------------------------------------------
    def remapIds(self,scenario):
        from pprint import pprint
        with VerboseGuard(f'Remapping piece ID for scenario '
                          f'"{scenario._name}"') as v:
            tmp = []
            for layout in scenario._layouts:
                cs = self._gamebox._sheets_map.get(layout._name,None)
                if not cs:
                    v(f'Missing counter sheet {layout._name} - maybe map')
                    continue
                
                tmp.extend(cs._piece+cs._card)


            # Map from ID to piece - only used counter sheets in this scenario
            # pprint(tmp)
            piece_map = {i: p for i,p in enumerate(tmp)}
            # print(piece_map)
            # piece_id  = {v: k for k,v in piece_map.items()}

        return piece_map
    
    # ----------------------------------------------------------------
    def changeSlot(self,slot,flipped,angle,changed):
        if angle == 0 and not flipped:
            return

        # Valid angles are
        #
        # for i in range(nangles):
        #    angle[i] = - i * 360 // nangles
        #
        # validAngles = [-i * 360 // nangles for i in range(nangles)]
        changed[slot['entryName']] = {'angle':angle,
                                      'step': 2 if flipped else 1}

    # ----------------------------------------------------------------
    def updateSlot(self,name,traits,changed):
        changes  = changed.get(name,None)
        if not changes:
            return

        with VerboseGuard(f'Updating rotation and step state of '
                          f'{name} to {changes}') as v:
            rot      = Trait.findTrait(traits, RotateTrait.ID)
            lyr      = Trait.findTrait(traits, LayerTrait.ID,
                                       key='name', value='Step')
            if not rot and not lyr:
                v(f'No traits to modify!')
                return
            
            angle    = changes['angle']
            step     = changes['step']
            nangle   = int(rot['nangles'])
            fangle   = angle
            if nangle != 1:
                # Funny calculation in VASSAL 
                a        = ((angle % 360) - 360) % 360
                e        = (-a / 360) * nangle
                fangle   = int(round(e) % nangle)
            # v(f'nangle={nangle} fangle={fangle}')
            if rot: rot.setState(angle = fangle)
            if lyr: lyr.setState(level = step)
        

    # ----------------------------------------------------------------
    def restoreSlot(self,slot,step,angle):
        # traits   = slot.getTraits()
        # rot      = Trait.findTrait(traits, RotateTrait.ID)
        # lyr      = Trait.findTrait(traits, LayerTrait.ID,
        #                            key='name', value='Step')
        # 
        # rot.setState(angle=rot)
        # lyr.setState(level=step)
        # 
        # slot.setTraits(*traits)
        pass 
    
    # ----------------------------------------------------------------
    def addScenario(self,scenario,menu):
        from sys import stderr
        from pprint import pprint
        
        with VerboseGuard(f'Adding a scenario {scenario._name}') as v:
            vsav    = VSav(self._build,self._vmod)
            save    = vsav.addSaveFile()
            changed = {}

            # ZunTzu reads in the counter sheets specified in a given
            # scenario, and then builds and array of those counters.
            # However, since we already have read in the counters, we
            # need to simply remap the index here.
            piece_map = self.remapIds(scenario)#self._gamebox._piece_map
            # pprint(piece_map)
            
            selectedBoard = None
            selectedMap   = None
            # First, loop over layouts and see if we can find the map
            # we're looking for.
            for layout in scenario._layouts:
                boardName = layout._name
                board     = self._game.getBoards(asdict=True).get(boardName,
                                                                  None)
                if not board:
                    # Try the next layout
                    continue

                map = board.getMap()
                if not map:
                    # A board without a map?
                    continue

                if not isinstance(map.getParent(),Game):
                    # Parent of map is not the game
                    continue

                selectedMap   = map
                selectedBoard = board
                break

            if not selectedMap:
                print(f'No map found for scenario {scenario._name}')

            mapName = selectedMap['mapName']
            boardName = selectedBoard['name']
            
            for layout in scenario._layouts:
                v(f'Layout {layout._name} with {len(layout._stacks)}')
                if len(layout._stacks) <= 0:
                    continue

                thisMap = mapName
                if self._allmaps:
                    boardName = layout._name
                    board     = self._game.getBoards(asdict=True)\
                                          .get(boardName, None)
                    
                    map = board.getMap() if board else None
                    if map:
                        # A board without a map?
                        thisMap = map['mapName']

                # In principle, it could be done like this, but to be
                # on the safe side, we do it a tad more laboriously.
                #
                # mapped = {(s._x,s._y):
                #           [self._piece_slots[self._gamebox._piece_map[i]]
                #            for i in s._ids]
                #           for s in layout._stacks}
                mapped  = {}
                v(f'Layout has {len(layout._stacks)}')
                for s in layout._stacks:
                    slots = []
                    for i,f,r in zip(s._ids,s._flip,s._rot):
                        v(f'Piece {i}')
                        piece = piece_map.get(i, None)
                        if not piece:
                            print(f'Could not find piece for id {i}',
                                  file=stderr)
                            continue
                        slot = self._piece_slots.get(piece, None)
                        if not slot:
                            print(f'Could not find slot for piece {piece}',
                                  file=stderr)
                            continue

                        # Set states
                        self.changeSlot(slot, f, r, changed)
                        
                        # v(f'id={i} -> piece={piece} -> slot={slot}')
                        slots.append(slot)

                    mapped[(int(s._x+.5),int(s._y+.5))] = slots

                # self._picker.getMap()['mapName']
                v(f'Add pieces to map: {thisMap}')
                save.addNoGrid(thisMap, mapped)

            v(f'Selected map/board {mapName}/{boardName}')
            
            picker = selectedMap.getBoardPicker(single=True)[0]
            
            if picker:
                picker.selectBoard(boardName)
                v(f'Selected map: {picker["selected"]}')
            
            # print('\n'.join(save.getLines()))
            # lines = []
            # save._otherLines(lines)
            # print('\n'.join(lines))
            vsav.run(savename    = scenario._file+'.vsav',
                     description = scenario._desc,
                     update      = lambda n,t : self.updateSlot(n,t,changed))

            if picker:
                picker.selectBoard(None)

            menu.addPredefinedSetup(name=scenario._name,
                                    useFile=True,
                                    file=scenario._file+'.vsav',
                                    description=scenario._desc)
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class PrototypeTrait(Trait):
    ID = 'prototype'
    def __init__(self,name=''):
        '''Create a prototype trait (VASSAL.counter.UsePrototype)'''
        super(PrototypeTrait,self).__init__()
        self.setType(name = name)
        self.setState(ignored = '')


Trait.known_traits.append(PrototypeTrait)

#
# EOF
#



## BEGIN_IMPORTS
from . base import DomInspect, ZTImage
from pywargame.common import VerboseGuard
## END_IMPORTS

# --------------------------------------------------------------------
class ZTMap(DomInspect):
    def __init__(self,elem):
        with VerboseGuard(f'Got a map'):
            self._name  = self._get_attr(elem,'name')
            self._image = ZTImage(elem)

    def read_image(self,zf):
        with VerboseGuard(f'Reading map ({self._name}) image') as v:
            self._image.read_image(zf)
            sc = ZTImage.target_dpi / self._image._reso
            v(f'Scale by {sc}={ZTImage.target_dpi}/{self._image._reso} '
              f'from {self._image._image.width}x{self._image._image.height}')
            self._image._image.resize(int(sc*self._image._image.width),
                                      int(sc*self._image._image.height))
            v(f'Scaled by {sc} to {self._image._image.width}x'
              f'{self._image._image.height}')
        
    @property
    def filename(self):
        return f'{self._name.replace(" ","_")}.{self._image._image.format.lower()}'

    @property
    def size(self):
        return self._image._image.width,self._image._image.height

    def __str__(self):
        return f' Map: name={self._name}'+'\n'+str(self._image)
#
# EOF
#



## BEGIN_IMPORT
from ... common import VerboseGuard
from .. trait import Trait
## END_IMPORT

# --------------------------------------------------------------------
class RestrictCommandsTrait(Trait):
    ID = 'hideCmd'
    HIDE = 'Hide'
    DISABLE = 'Disable'
    def __init__(self,
                 name          = '',
                 hideOrDisable = HIDE,
                 expression    = '',# Restrict when true
                 keys          = []):
        '''Create a layer trait (VASSAL.counter.RestrictCommands)'''
        super(RestrictCommandsTrait,self).__init__()
        encKeys = ','.join([k.replace(',',r'\,') for k in keys])
        self.setType(name          = name,
                     hideOrDisable = hideOrDisable,
                     expression    = expression,
                     keys          = encKeys)
        self.setState(state='')
    def setKeys(self,keys):
        self['keys'] = ','.join([k.replace(',',r'\,') for k in keys])
    

Trait.known_traits.append(RestrictCommandsTrait)

#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . archive import Archive
from . features import Features
## END_IMPORT

def num_version(major,minor):
    return major * 256 + minor

def readVector(ar,cls):
    with VerboseGuard('Reading vector') as g:
        if Features().size_size == 8:
            n = ar.size()
        else:
            n                   = ar.word()
            if n == 0xFFFF:
                n               = ar.dword()
                if n == 0xFFFFFFFF:
                    n           = ar.int(64)
        g(f'{n} elements')
    return [cls(ar) for _ in range(n)]

# ====================================================================
class GBXHeader:
    BOX      = 'GBOX'
    SCENARIO = 'GSCN'
    def __init__(self,ar,expect=BOX):
        '''GBXHeader of file 

        4 bytes format ID
        4x1 byte format and program version

        8 bytes in total 
        '''
        with VerboseGuard('Reading header') as g:
            sig = ar.chr(len(expect))
            assert sig == expect, f'Not a {expect} file: {sig}'

            self._major        = ar.byte()
            self._minor        = ar.byte()
            self._programMajor = ar.byte()
            self._programMinor = ar.byte()
            self._vers         = num_version(self._major,self._minor)
            g(f'Version {self._major}.{self._minor}')

            assert self._vers >= num_version(3,0),\
                f'{self._major}.{self._minor} format not supported'

            if self._vers >= num_version(4,0):
                g(f'Detected version 4.0 or newer, setting some features')
                Features().id_size       = 4
                Features().size_size     = 8
                Features().sub_size      = 8
                Features().square_cells  = True
                Features().rotate_unit   = True
                Features().piece_100     = True
                Features().private_board = True
                Features().roll_state    = True
                
                
            


    def __str__(self):
        return ('Header:\n'
                f'  Format major version:  {self._major}\n'
                f'  Format minor version:  {self._minor}\n'
                f'  Program major version: {self._programMajor}\n'
                f'  Program minor version: {self._programMinor}\n')


# --------------------------------------------------------------------
class GBXStrings:
    def __init__(self,ar):
        '''Map IDs to strings'''
        with VerboseGuard(f'Reading string mappings'):
            strMapN = ar.size()
            
            self._id2str = {}
            for _ in range(strMapN):
                key = ar.dword()
                val = ar.str()
            
                self._id2str[key] = val

    def __str__(self):
        return ('Strings:\n'+
                '\n'.join([f'  {key:8x}: {val}'
                           for key,val in self._id2str.items()]))

# --------------------------------------------------------------------
class GSNStrings:
    def __init__(self,ar):
        '''Map IDs to strings'''
        with VerboseGuard(f'Reading string mappings'):
            strMapN = ar.size()
            
            self._id2str = {}
            for _ in range(strMapN):
                key = ar.size()
                val = ar.str()
            
                self._id2str[key] = val

    def __str__(self):
        return ('Strings:\n'+
                '\n'.join([f'  {key:8x}: {val}'
                           for key,val in self._id2str.items()]))
    
#
# EOF
#



## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
## END_IMPORT

class BaseFolder(Element):
    UNIQUE = ['name']
    
    def __init__(self,parent,tag,node=None,name='',description='',**kwargs):
        '''Create a folder'''
        super().__init__(parent,
                         tag,
                         node = node,
                         name = name,
                         desc = description,
                         **kwargs)

# --------------------------------------------------------------------
class GlobalPropertyFolder(BaseFolder):
    TAG = Element.FOLDER+'GlobalPropertySubFolder'
    def __init__(self,
                 parent,
                 node=None,
                 name='',
                 description=''):
        super().__init__(parent,
                         tag = self.TAG,
                         node = node,
                         name = name,
                         description = description)

registerElement(GlobalPropertyFolder)
    
# --------------------------------------------------------------------
class DeckFolder(BaseFolder):
    TAG = Element.FOLDER+'DeckSubFolder'
    def __init__(self,
                 parent,
                 node=None,
                 name='',
                 description=''):
        super().__init__(parent,
                         tag = self.TAG,
                         node = node,
                         name = name,
                         description = description)

registerElement(DeckFolder)
        
# --------------------------------------------------------------------
class MapFolder(BaseFolder):
    TAG = Element.FOLDER+'MapSubFolder'
    def __init__(self,
                 parent,
                 node=None,
                 name='',
                 description=''):
        super().__init__(parent,
                         tag = self.TAG,
                         node = node,
                         name = name,
                         description = description)
        
registerElement(MapFolder)

# --------------------------------------------------------------------
class ModuleFolder(BaseFolder):
    TAG = Element.FOLDER+'ModuleSubFolder'
    def __init__(self,
                 parent,
                 node=None,
                 name='',
                 description=''):
        super().__init__(parent,
                         tag = self.TAG,
                         node = node,
                         name = name,
                         description = description)
        
registerElement(ModuleFolder)

# --------------------------------------------------------------------
class PrototypeFolder(BaseFolder):
    TAG = Element.FOLDER+'PrototypeSubFolder'
    def __init__(self,
                 parent,
                 node=None,
                 name='',
                 description=''):
        super().__init__(parent,
                         tag = self.TAG,
                         node = node,
                         name = name,
                         description = description)
        
registerElement(PrototypeFolder)

# --------------------------------------------------------------------
#
# EOF
#




## BEGIN_IMPORT
from .. common import VerboseGuard
from . base import *
from . element import Element
from . gameelements import GameElement
from . mapelements import MapElement
from . zone import *
from . grid import *
## END_IMPORT

# --------------------------------------------------------------------
class BoardPicker(MapElement):
    TAG = Element.MAP+'BoardPicker'
    def __init__(self,doc,node=None,
                 addColumnText        = 'Add column',
                 addRowText           = 'Add row',
                 boardPrompt          = 'Select board',
                 slotHeight           = 125,
                 slotScale            = 0.2,
                 slotWidth            = 350,
                 title                = 'Choose Boards'):
        super(BoardPicker,self).__init__(doc,self.TAG,node=node,
                                         addColumnText        = addColumnText,
                                         addRowText           = addRowText,
                                         boardPrompt          = boardPrompt,
                                         slotHeight           = slotHeight,
                                         slotScale            = slotScale,
                                         slotWidth            = slotWidth,
                                         title                = title,
                                         selected             = '')

    def addSetup(self,**kwargs):
        '''Add a `Setup` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Setup
            The added element
        '''
        if 'mapName' not in kwargs:
            m = self.getMap()
            kwargs['mapName'] = m.getAttribute('mapName')
            
        return self.add(Setup,**kwargs)
    def getSetups(self,single=False):
        '''Get all or a sole `Setup` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `Setup` child, otherwise fail.
            If `False` return all `Setup` children in this element
        
        Returns
        -------
        children : list
            List of `Setup` children (even if `single=True`)
        '''
        return self.getAllElements(Setup,single=single)
    def addBoard(self,**kwargs):
        '''Add a `Board` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : Board
            The added element
        '''
        return self.add(Board,**kwargs)
    def getBoards(self,asdict=True):
        '''Get all Board element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Board` elements.  If `False`, return a list of all Board` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Board` children
        '''
        return self.getElementsByKey(Board,'name',asdict=asdict)
    
    def selectBoard(self,name):
        if name is None:
            self.setAttribute('selected','')
            return
        
        if name not in self.getBoards():
            print(f'Board "{name}" not in "{self.getMap()["mapName"]}" picker')
            return

        escname = name.replace('|',' ')
        self.setAttribute('selected',f'{self["selected"]}|{name}|')
        #print(f'Added "{name}" to selected boards: {self["selected"]}')
        
    def encode(self):
        setups = self.getSetups()
        if setups is not None and len(setups)>0:
            return [setups[0]._node.childNodes[0].nodeValue]
        
        ret    = []
        selected = self['selected']
        #print(f'Selected boards: {selected}')
        for bn in self.getBoards().keys():
            escname = '|'+bn.replace('|',' ')+'|'
            # if selected != '':
            #     print(f'Ignore board "{bn}" in map '
            #           f'{self.getMap()["mapName"]} '
            #           f'"{selected}" -> '
            #           f'{escname not in selected}')
            if escname not in selected:
                continue 
            ret.append(self.getMap()['mapName']+'BoardPicker\t'+bn+'\t0\t0')

        return ret

registerElement(BoardPicker)

# --------------------------------------------------------------------
class Setup(Element):
    TAG = 'setup'
    def __init__(self,picker,node=None,
                 mapName = '',
                 maxColumns = 1,
                 boardNames = []):
        super(Setup,self).__init__(picker,self.TAG,node=node)
        col = 0
        row = 0
        lst = [f'{mapName}BoardPicker']
        for bn in boardNames:
            lst.extend([bn,str(col),str(row)])
            col += 1
            if col >= maxColumns:
                col = 0
                row += 1
                
        txt = r'	'.join(lst)
        self.addText(txt)

    def getPicker(self): return self.getParent(BoardPicker)

registerElement(Setup)
    
# --------------------------------------------------------------------
class Board(Element):
    TAG = Element.PICKER+'Board'
    UNIQUE = ['name']
    def __init__(self,picker,node=None,
                 name       = '',
                 image      = '',
                 reversible = False,
                 color      = rgb(255,255,255),
                 width      = 0,
                 height     = 0):
        super(Board,self).__init__(picker,self.TAG,node=node,
                                   image      = image,
                                   name       = name,
                                   reversible = reversible,
                                   color      = color,
                                   width      = width,
                                   height     = height)

    def getPicker(self): return self.getParent(BoardPicker)
    def getMap(self):
        z = self.getPicker()
        if z is not None:
            return z.getMap()
        return None
    def addZonedGrid(self,**kwargs):
        '''Add a `ZonedGrid` element to this

        Parameters
        ----------
        kwargs : dict
            Dictionary of attribute key-value pairs
        
        Returns
        -------
        element : ZonedGrid
            The added element
        '''
        return self.add(ZonedGrid,**kwargs)
    def getZonedGrids(self,single=True):
        '''Get all or a sole `ZonedGrid` element(s) from this

        Parameters
        ----------
        single : bool
            If `True`, there can be only one `ZonedGrid` child, otherwise fail.
            If `False` return all `ZonedGrid` children in this element
        
        Returns
        -------
        children : list
            List of `ZonedGrid` children (even if `single=True`)
        '''
        return self.getAllElements(ZonedGrid,single=single)
    def getZones(self,asdict=True):
        '''Get all Zone element(s) from this

        Parameters
        ----------
        asdict : bool
            If `True`, return a dictonary that maps key to `Zone` elements.  If `False`, return a list of all Zone` children.
        
        Returns
        -------
        children : dict or list
            Dictionary or list of `Zone` children
        '''
        zoned = self.getZonedGrids(single=True)
        if zoned is None: return None

        return zoned[0].getZones(asdict=asdict)

    def getWidth(self):
        # print(f'Getting width of {self}: {self["width"]}')
        if 'width' in self and int(self['width']) != 0:
            return int(self['width'])
        return 0

    def getHeight(self):
        # print(f'Getting height of {self}: {self["height"]}')
        if 'height' in self and int(self['height']) != 0:
            return int(self['height'])
        return 0

registerElement(Board)

#
# EOF
#
